/* sensor driver for BeOS for Abit BP6 board
 *
 * R4.52 port by Robert Pointon <b.g.fish@altavista.net>
 *	sensor chip = WinBond W83782D
 *  - see following sources:
 */

/*===========================================================================*/
/*
 * LM78 sensor driver for BeOS 
 *	 R4 port by Andrew Vessey (avessey@mindspring.com)
 *		- fixed LM78_get sprintf prob
 *		- moved code to R4 style bus manager read/writes
 *   original R3 port by Kazuki Sakamoto (sakamoto@cec.co.jp)
 */
/*===========================================================================*/ 
/* -*- linux-c -*-
 *
 *  LM78 sensor driver for LINUX 
 *
 * Copyright (c) 1997 Alexander Larsson
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version
 *  2 of the License, or (at your option) any later version.
 * 
 * provides a /proc/sensors which shows
 * some information about your system.
 * Voltages, mb temperature and fan speeds
 * are currently reported.
 *
 * This should work on all ASUS TX97 cards
 * but it's only tested on a TX97-X.
 * It should also work on any card that has
 * a LM78 circuit, but some constants in
 * lm78.h probably has to be changed.
 *
 * please report any bugs/patches to me:
 *  alla@lysator.liu.se
 *
 * Reference:
 *   National Semiconductor
 *   LM78 Microprocessor System Hardware Monitor, September 1996.
 *   http://www.national.com/ds/LM/LM78.pdf
 *
 */
/*===========================================================================*/
/*------------------------------------------------------------------
  lm78.c

  A Linux driver for National Semiconductor LM78 Microprocessor System
  Hardware Monitor.  (data sheet available from http://www.national.com)

  Erik Hendriks
  hendriks@cesdis.gsfc.nasa.gov

  $Id: lm78.c,v 1.4 1998/08/17 15:09:39 hendriks Exp $
  ------------------------------------------------------------------*/
/*===========================================================================*/  
/*
    w83781d.c - Part of lm_sensors, Linux kernel modules for hardware
                monitoring
    Copyright (c) 1998, 1999  Frodo Looijaard <frodol@dds.nl>,
    Philip Edelbrock <phil@netroedge.com>,
    and Mark Studebaker <mds@eng.paradyne.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
/*===========================================================================*/
#include <KernelExport.h>
#include <Drivers.h>
#include <Errors.h>
#include <ISA.h>
#include <stdio.h>

#define LM78_BASE		0x290
#define LM78_ADDRESS_REG	(LM78_BASE + 0x5)
#define LM78_DATA_REG		(LM78_BASE + 0x6)

#define	LM78_WAIT		1000	/* 1000us */

#define LM78_BUSY		0x80
#define LM78_TEST		0x20

#define LM78_IN0 0x20     
#define LM78_IN1 0x21     
#define LM78_IN2 0x22     
#define LM78_IN3 0x23     
#define LM78_IN4 0x24     
#define LM78_IN5 0x25
#define LM78_IN6 0x26      
#define LM78_TEMP 0x27
#define LM78_FAN0 0x28
#define LM78_FAN1 0x29
#define LM78_FAN2 0x2A

#define LM78_FANDIV 0x47

const char *driver_name="lm78";

static const char *name_list[] = {
	"misc/lm78",
	NULL
};

static status_t lm78_open(const char *, uint32, void **);
static status_t lm78_close(void *);
static status_t lm78_free(void *);
static status_t lm78_control(void *, uint32, void *, size_t);
static status_t lm78_read(void *, off_t, void *, size_t *);
static status_t lm78_write(void *, off_t, const void *, size_t *); 

int32 api_version = B_CUR_DRIVER_API_VERSION;
static long is_open;
isa_module_info * isa;

/*--- These are specific to the Winbond W83781D ---*/
#define WB_BANK_SEL        0x4E
/* The W83781D has a notion of register banks.  To
 * denote which bank a register exists in, I'm sticking
 * the bank number in bits 8-15 of the register name */
#define WB_IN7				0x0551
#define WB_IN8				0x0550
#define WB_DATA_TEMP1       0x0150
#define WB_DATA_TEMP2       0x0250

#define WB_WCHIPID 			0x58

static inline void
LM78_select(uint regno) 
{	while (isa->read_io_8(LM78_ADDRESS_REG) & LM78_BUSY)
		spin(LM78_WAIT);
    isa->write_io_8(LM78_ADDRESS_REG,regno);
}

static inline uint8
LM78_get(uint regno) {
    LM78_select(regno);
    return isa->read_io_8(LM78_DATA_REG);
}

static inline void
LM78_set(uint regno, uint8 value) {
    LM78_select(regno);
    isa->write_io_8(LM78_DATA_REG,value);
}

static inline void
WB_bank_select(uint bank) {
    uint8 b;
    b = LM78_get(WB_BANK_SEL);
    b &= ~0x87;
    b |= bank;
    LM78_set(WB_BANK_SEL, b);
}

static inline uint8
WB_get(uint reg) {
    uint8 val;
    if (reg & 0xFF00) {
	WB_bank_select(reg>>8);
	val = LM78_get(reg&0xFF);
	WB_bank_select(0);
    } else
	val = LM78_get(reg&0xFF);
    return val;
}

_EXPORT status_t
init_hardware(void)
{	uint8 ret;
	dprintf("%s: init_hardware()\n",driver_name);	
	if (get_module(B_ISA_MODULE_NAME, (module_info **) &isa) <0)
		return ENOSYS;
	/* Verify that lm78 type architecture exists */
	isa->write_io_8(LM78_ADDRESS_REG, LM78_TEST);
	ret = isa->read_io_8(LM78_ADDRESS_REG);
	isa->read_io_8(LM78_DATA_REG);
	if ((ret & 0x7f) != LM78_TEST)
		return  B_ERROR;	
	/* Set FAN1 and FAN2 to divide by 2 */
	LM78_set(LM78_FANDIV, 0x50);
	put_module(B_ISA_MODULE_NAME);
	return  B_OK;
}

_EXPORT status_t
init_driver(void)
{	dprintf("%s: init_driver()\n",driver_name);
	if (get_module(B_ISA_MODULE_NAME, (module_info **) &isa) <0)
		return ENOSYS;	
	is_open = 0;
	return B_OK;
	
}

_EXPORT void
uninit_driver(void)
{	dprintf("%s: uninit_driver()\n",driver_name);
	put_module(B_ISA_MODULE_NAME);
}

static status_t
lm78_open(const char *name, uint32 flags, void **cookie)
{	int id;
	dprintf("%s: open(%s)\n",driver_name,name);
	id = lookup_device_name(name);
	if (id < 0){
		return EINVAL;}

	if (atomic_or(&is_open, 1) & 1)
	{
		return B_BUSY;}
	return B_OK;
}

static status_t
lm78_close(void *cookie)
{	dprintf("%s: close()\n",driver_name);
	return B_OK;
}

static status_t
lm78_free(void *cookie)
{	dprintf("%s: free()\n",driver_name);
	atomic_and(&is_open, ~1);
	return B_OK;
}

static status_t
lm78_control(void *cookie, uint32 op, void *data, size_t len)
{	dprintf("%s:control(%d)\n",driver_name,op);
	switch (op) {
	default:
		return EINVAL;
	}
	return B_OK;
}

static status_t
lm78_write(void *cookie, off_t pos, const void *data, size_t *len)
{	dprintf("%s: write()\n",driver_name);
	*len = 0;
	return B_IO_ERROR;
}

static void myprintf(void *buff,size_t *len,char *format,uint8 val)
{	sprintf((char*)buff + *len, format, val);
	*len = strlen((char*)buff);
}

static status_t
lm78_read(void *cookie, off_t pos, void *data, size_t *len)
{	int wb;
	dprintf("%s: read()\n",driver_name);
	*len = 0; 	
	/* Determine if and what WinBond type*/
	wb=1;
	switch(LM78_get(WB_WCHIPID) & 0xfe)	/* mask off lower bit, not reliable */
	{	case 0x10:myprintf(data, len, "w83781d\n",0);break;
		case 0x30:myprintf(data, len, "w83782d\n",0);break;  /*ABIT BP6 uses this*/
		case 0x40:myprintf(data, len, "w83783s\n",0);break;
		default:  myprintf(data, len, "lm78\n",0);wb=0;/*a plain lm78?*/
	}			
	/* Like the 'Fans' we return 255 if something isn't supported*/
    myprintf(data, len, "IN0: %d\n",LM78_get(LM78_IN0));
	myprintf(data, len, "IN1: %d\n", LM78_get(LM78_IN1)); 
	myprintf(data, len, "IN2: %d\n", LM78_get(LM78_IN2));
	myprintf(data, len, "IN3: %d\n", LM78_get(LM78_IN3));
	myprintf(data, len, "IN4: %d\n", LM78_get(LM78_IN4));
	myprintf(data, len, "IN5: %d\n", LM78_get(LM78_IN5));
	myprintf(data, len, "IN6: %d\n", LM78_get(LM78_IN6));
	myprintf(data, len, "IN7: %d\n", (wb==0)?255:WB_get(WB_IN7));
	myprintf(data, len, "IN8: %d\n", (wb==0)?255:WB_get(WB_IN8));
	myprintf(data, len, "TEMP0: %d\n", LM78_get(LM78_TEMP));
	myprintf(data, len, "TEMP1: %d\n", (wb==0)?255:WB_get(WB_DATA_TEMP1));
	myprintf(data, len, "TEMP2: %d\n", (wb==0)?255:WB_get(WB_DATA_TEMP2));
	myprintf(data, len, "FAN0: %d\n", LM78_get(LM78_FAN0));
	myprintf(data, len, "FAN1: %d\n", LM78_get(LM78_FAN1));
	myprintf(data, len, "FAN2: %d\n", LM78_get(LM78_FAN2));
	return B_OK;
}

_EXPORT static device_hooks lm78_hooks = {
	lm78_open,
	lm78_close,
	lm78_free,
	lm78_control,
	lm78_read,
	lm78_write
};

_EXPORT const char**
publish_devices()
{
	return name_list;
}

int
lookup_device_name(const char *name)
{	
	int i;
	for (i = 0; name_list[i]; i++)
		if (!strcmp (name_list[i], name))
			{
			return i;}
	return -1;		/* Invalid device name */
}

_EXPORT device_hooks *
find_device(const char *name)
{
	if (lookup_device_name(name) >= 0)
		{
		return &lm78_hooks;	/* Return hooks list */
		}
	return NULL;		/* Device not found */
}








	








