/*------------------------------------------------------------------
  lm78.c

  A Linux driver for National Semiconductor LM78 Microprocessor System
  Hardware Monitor.  (data sheet available from http://www.national.com)

  Erik Hendriks
  hendriks@cesdis.gsfc.nasa.gov

  $Id: lm78.c,v 1.4 1998/08/17 15:09:39 hendriks Exp $
  ------------------------------------------------------------------*/
#include <linux/version.h>
#if LINUX_VERSION_CODE < 0x20100
#define VERSION_2_0 1
#endif

#include <linux/config.h>
#ifdef CONFIG_MODVERSIONS
#define MODVERSIONS 1
#include <linux/modversions.h>
#endif

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/malloc.h>
#include <linux/types.h>
#include <linux/timer.h>
#include <linux/ctype.h>
#include <linux/sysctl.h>
#include <linux/ioport.h>
#define REALLY_SLOW_IO 1
#include <asm/io.h>
#if (! VERSION_2_0)
#include <asm/uaccess.h>
#endif

#define LM78_BASE io
#include <lm78.h>
#include <monitor.h>

static char *idstring="$Id: lm78.c,v 1.4 1998/08/17 15:09:39 hendriks Exp $";
int LM78_BASE = LM78_DEF_BASE;
int monitor_type = 0;

#ifndef VERSION_2_0
MODULE_AUTHOR("Erik Hendriks <hendriks@cesdis.gsfc.nasa.gov>");
MODULE_DESCRIPTION("Driver for National Semiconductor LM78 type hardware monitors.");
MODULE_PARM(io, "i");
MODULE_PARM_DESC(io, "Base address of hardware monitor.");
MODULE_PARM(monitor_type, "i");
MODULE_PARM_DESC(monitor_type, "Hardware monitor type (0=autodetect)");
#endif

static int last_check=0;
static int last_ints=0;

#if VERSION_2_0
int proc_dointvec(ctl_table *table, int write, struct file *filp, void *buffer, size_t *lenp);
int proc_dostring(ctl_table *table, int write, struct file *filp, void *buffer, size_t *lenp);
#endif

static int detect_monitor(void) {
    /* Doc errata: My W83781D returns 0x11 in register 0x58, bank 0 */
    if (WB_get(WB_CHIP_ID)          == 0x11 && 
	WB_read_short(WB_VENDOR_ID) == 0x5ca3) {
	/* Perform a reset for this guy */
	LM78_set(LM78_CONFIG, LM78_CONF_INIT);
	if (LM78_get(LM78_CONFIG) == WB_CONF_INIT) {
	    printk("LM78: Detected a Winbond W83781D at 0x%x\n", LM78_BASE);
	    return W83781D;
	}
    }

    /* Check for a LM78-J */
    if (LM78_get(LM78_CHIP_ID) & 0x40) {
	/*LM78_set(LM78_CHIP_ID, 0x20);*/
	if (LM78_get(LM78_CONFIG) == LM78_CONF_INT) {
	    printk("LM78: Detected LM78-J at 0x%x\n", LM78_BASE);
	    return LM78J;
	}
    }
    
    /* Finally, check for older LM78 (no chip ID here) */
    LM78_set(LM78_CONFIG, LM78_CONF_INIT);
    if (LM78_get(LM78_CONFIG) == LM78_CONF_INT) {
	printk("LM78: Detected LM78 at 0x%x\n", LM78_BASE);
	return LM78;
    }    
    
    return 0;
}


static void sysctl_add_item (int id, int reg, char *name, int mode, int flags);

#define INT  SC_TYPE_INT|sizeof(int)
#define INT2 SC_TYPE_INT|sizeof(int)*2
static void
setup_sysctl(void) {
    /*--- Stuff that every monitor has ---*/
    sysctl_add_item(HM_IN0, LM78_LIM_IN0,      "in0", 0644, SC_LIMIT |INT*2);
    sysctl_add_item(HM_IN0, LM78_DATA_IN0,     "in0", 0444, SC_STATUS|INT);
    sysctl_add_item(HM_IN1, LM78_LIM_IN1,      "in1", 0644, SC_LIMIT |INT*2);
    sysctl_add_item(HM_IN1, LM78_DATA_IN1,     "in1", 0444, SC_STATUS|INT);
    sysctl_add_item(HM_IN2, LM78_LIM_IN2,      "in2", 0644, SC_LIMIT |INT*2);
    sysctl_add_item(HM_IN2, LM78_DATA_IN2,     "in2", 0444, SC_STATUS|INT);
    sysctl_add_item(HM_IN3, LM78_LIM_IN3,      "in3", 0644, SC_LIMIT |INT*2);
    sysctl_add_item(HM_IN3, LM78_DATA_IN3,     "in3", 0444, SC_STATUS|INT);
    sysctl_add_item(HM_IN4, LM78_LIM_IN4,      "in4", 0644, SC_LIMIT |INT*2);
    sysctl_add_item(HM_IN4, LM78_DATA_IN4,     "in4", 0444, SC_STATUS|INT);
    sysctl_add_item(HM_IN5, LM78_LIM_IN5,      "in5", 0644, SC_LIMIT |INT*2);
    sysctl_add_item(HM_IN5, LM78_DATA_IN5,     "in5", 0444, SC_STATUS|INT);
    sysctl_add_item(HM_IN6, LM78_LIM_IN6,      "in6", 0644, SC_LIMIT |INT*2);
    sysctl_add_item(HM_IN6, LM78_DATA_IN6,     "in6", 0444, SC_STATUS|INT);

    /*--- For backward compatibility? ---*/
    sysctl_add_item(HM_TEMP0, LM78_LIM_OVERTEMP, "temperature", 0644, SC_LIMIT |INT*2);
    sysctl_add_item(HM_TEMP0, LM78_DATA_TEMP,    "temperature", 0444, SC_STATUS|INT);
    /*-----------------------------------*/

    sysctl_add_item(HM_TEMP0, LM78_LIM_OVERTEMP, "temp0", 0644, SC_LIMIT |INT*2);
    sysctl_add_item(HM_TEMP0, LM78_DATA_TEMP,    "temp0", 0444, SC_STATUS|INT);
    sysctl_add_item(HM_FAN0,  LM78_LIM_FAN1,     "fan0", 0644, SC_LIMIT |INT*2);
    sysctl_add_item(HM_FAN0,  LM78_DATA_FAN1,    "fan0", 0444, SC_STATUS|INT);
    sysctl_add_item(HM_FAN1,  LM78_LIM_FAN2,     "fan1", 0644, SC_LIMIT |INT*2);
    sysctl_add_item(HM_FAN1,  LM78_DATA_FAN2,    "fan1", 0444, SC_STATUS|INT);
    sysctl_add_item(HM_FAN2,  LM78_LIM_FAN3,     "fan2", 0644, SC_LIMIT |INT*2);
    sysctl_add_item(HM_FAN2,  LM78_DATA_FAN3,    "fan2", 0444, SC_STATUS|INT);

    sysctl_add_item(HM_CHASSIS, LM78_DATA_CHASSIS,    "chassis", 0644, SC_STATUS|INT);
    sysctl_add_item(HM_INTERRUPTS, LM78_DATA_INTERRUPTS, "interrupts", 0644, SC_STATUS|SC_TYPE_STR|30);

    sysctl_add_item(HM_RESET, LM78_CONTROL_RESET, "reset", 0200, SC_CONTROL|SC_TYPE_STR|1);

    if (monitor_type == W83781D) {
	sysctl_add_item(HM_TEMP1, WB_LIM_HYST_2,  "temp1", 0644, SC_LIMIT |INT*2);
	sysctl_add_item(HM_TEMP1, WB_DATA_TEMP_2, "temp1", 0444, SC_STATUS|INT);
	sysctl_add_item(HM_TEMP2, WB_LIM_HYST_3,  "temp2", 0644, SC_LIMIT |INT*2);
	sysctl_add_item(HM_TEMP2, WB_DATA_TEMP_3, "temp2", 0444, SC_STATUS|INT);
	/*sysctl_add_item(HM_BEEP,  WB_BEEP_CTRL,   "beep",  0644, SC_CONTROL|INT);*/
    }
}

static int
LM78_initialize(void) {
    unsigned char c;

    if (check_region(LM78_ADDR, 2)) {
	 printk("Ports 0x%x-0x%x are used by another driver.\n",
		(int)LM78_ADDR, (int)LM78_ADDR+1);
	 return -EBUSY;
    }

    /* Generic cleanup to help catch chips in weird states. */
    LM78_set(LM78_CONFIG, LM78_CONF_INIT);/* Reset */
    WB_bank_select(0);

    if (!monitor_type) {
	monitor_type = detect_monitor();
	if (monitor_type == 0) {
	    printk("LM78: Detection failed at 0x%x\n", LM78_BASE);
	    return -EINVAL;
	}
    }

    /* Grab the IO ports that we need */
    request_region(LM78_ADDR, 2, "LM78");

    /* Mask off all interrupts by default. */
    LM78_set(LM78_CONF_SMI1, 0xFF);
    LM78_set(LM78_CONF_SMI2, 0xFF);
    LM78_set(LM78_CONF_NMI1, 0xFF);
    LM78_set(LM78_CONF_NMI2, 0xFF);

    /* Start the LM78 monitoring.
     * Set   START
     * Clear INT__CLear
     */
    c = LM78_get(LM78_CONFIG);
    c |=  LM78_CONF_START;	/* Start monitoring */
    c &= ~LM78_CONF_INT;	/* Allow interrupts  */

    c |=  LM78_CONF_NMI;	/* Select NMI/IRQ interrupt reporting */
    c &= ~LM78_CONF_IRQ;	/* Select IRQ interrupt reporting */
    LM78_set(LM78_CONFIG, c);

    return 0;
}

static void LM78_stop(void) {
    LM78_and_reg(LM78_CONFIG, ~LM78_CONF_START);
    release_region(LM78_ADDR, 2);
}

/*--------------------------------------------------------------------
 *  LM78_get_interrupts
 */
static unsigned short LM78_get_interrupts(void) {
    unsigned char t1, t2;
    if (jiffies > last_check+(1.5*HZ)) {
	t1 = LM78_get(LM78_DATA_INT1);
	t2 = LM78_get(LM78_DATA_INT2);
	last_check = jiffies;
	last_ints = ((t2 << 8) | t1);
    }
    return last_ints;
}

static void LM78_flush_interrupts(void) {
    LM78_get(LM78_DATA_INT1);
    LM78_get(LM78_DATA_INT2);
    last_check = jiffies;
    last_ints = 0;
}

/*--------------------------------------------------------------------
 *  LM78_clear_chassis
 *
 * The chassis intrusion flag doesn't get reset by reading interrupts
 * like the others.
 */
static void LM78_clear_chassis(void) {
    unsigned char c;
    c = LM78_get(LM78_CONF_NMI2);
    c |= 0x80;
    LM78_set(LM78_CONF_NMI2, c);
    /* Clear that bit from the saved interrupts too */
    last_ints &= ~LM78_INT_CHASSIS;
}


/*--------------------------------------------------------------------
 *  Fan functions.
 */
static int LM78_get_fan_divisor(int fan) {
    int v;
    v = LM78_get(LM78_CONF_VID_FAN);
    switch(fan) {
    case 1: return 1 << ((v >> 4) & 0x3);
    case 2: return 1 << ((v >> 6) & 0x3);
    case 3: return 2;
    }
    printk("Asked for invalid fan number: %d\n", fan);
    return 0;
}

static int LM78_set_fan_divisor(int fan, int divisor) {
    int c, v;
    switch(divisor) {
    case 1: v = 0; break;
    case 2: v = 1; break;
    case 4: v = 2; break;
    case 8: v = 3; break;
    default:
	printk("Invalid fan divisor: %d\n", divisor);
	return -1;
    }
    c = LM78_get(LM78_CONF_VID_FAN);
    switch(fan) {
    case 1:
	c &=  ~(3 << 4);
	c |=   v << 4;
	LM78_set(LM78_CONF_VID_FAN, c);
	break;
    case 2:
	c &=  ~(3 << 6);
	c |=   v << 6;
	LM78_set(LM78_CONF_VID_FAN, c);
	break;
    case 3:
	if (divisor != 2) {
	    printk("Can't set divisor for fan 3\n");
	    return -1;
	}
	break;
    default:
	printk("Invalid fan number: %d\n", fan);
	return -1;
    }
    return 0;
}

static inline
int LM78_count_to_speed(int ct, int divisor) {
    if (ct == 255) return 0;	/* Stopped fans should read zero */
    if (ct == 0)   ct = 1;	/* Let's not divide by zero */
    return 1350000/(ct * divisor);
}

static inline
int LM78_speed_to_count(int rpm, int divisor) {
    int v;
    if (rpm == 0) return 255;	/* Let's not divide by zero */
    v = 1350000/(rpm * divisor);
    if (v > 255) v = 255;
    return v;
}

static
int LM78_get_fan_speed(int fan) {
    int ct, div;

    ct = LM78_get(LM78_DATA_FAN1+fan-1);
    div = LM78_get_fan_divisor(fan);
    return LM78_count_to_speed(ct, div);
}

/*--------------------------------------------------------------------
 *  LM78_hardware_reset
 */   
static void
LM78_hardware_reset(void) {
    LM78_or_reg(LM78_CONF_SMI2, 0x80);
    LM78_or_reg(LM78_CONFIG, LM78_CONF_RESET);
    printk("LM78: Reset failed.\n");
}

static char *lm78_int_names[] = LM78_interrupt_names;
static void lm78_print_ints(char *dest) {
    int i, ints;
    ints = LM78_get_interrupts();
    dest[0]=0;
    for (i=0; i < 16; i++)
	if (ints & (1<<i)) {
	    strcat(dest, lm78_int_names[i]);
	    strcat(dest, " ");
	}
}

/*--------------------------------------------------------------------
 *--------------------------------------------------------------------
 *--------------------------------------------------------------------
 *  
 */
/* Macros to simplify the data access */
#define DATA_int(n) (((int*)data)[(n)])
#define DATA_string ((char *)data)
static int
LM78_set_item(int reg, void *data) {
    int regval;

    /*--- write ----------------------------------------------------*/
    switch (reg) {
/*--- LIMIT items ---*/
#define low_speed (DATA_int(0))
#define divisor   (DATA_int(1))
    case LM78_LIM_FAN1:	/* These have 1 limit value */
    case LM78_LIM_FAN2:	/* Only "high" is limit in RPM */
    case LM78_LIM_FAN3:	/* low = fan speed divisor */
	/* Check that the divisor is valid */
	if (low_speed < 0) return -EINVAL;
	if (divisor != 1 && divisor != 2 && divisor !=4 && divisor != 8) return -EINVAL;
	if (reg == LM78_LIM_FAN3 && divisor != 2) return -EINVAL;

	regval = LM78_speed_to_count(low_speed, divisor);

	/* The the fan and the divisor */
	LM78_set_fan_divisor(reg-LM78_LIM_FAN1+1, divisor);
	LM78_set(reg, regval);
	break;
#undef low_speed
#undef divisor
#define high (DATA_int(0))
#define low  (DATA_int(1))
    case LM78_LIM_OVERTEMP:
	/* Temp has signed values... */
	if (high < -128 || high > 127 ||
	    low  < -128 || low  > 127) return -EINVAL;
	if (high < low) return -EINVAL;
	LM78_set(reg,   high);
	LM78_set(reg+1, low);
	break;
    case WB_LIM_HYST_2:
    case WB_LIM_HYST_3:
	if (high < -128 || high > 127 ||
	    low  < -128 || low  > 127) return -EINVAL;
	if (high < low) low = 127; /* Set to comparator mode */
	WB_set(reg,   low);
	WB_set(reg+2, high);
	break;
    case LM78_LIM_IN0:/* Everything else has high/low */
    case LM78_LIM_IN1:
    case LM78_LIM_IN2:
    case LM78_LIM_IN3:
    case LM78_LIM_IN4:
    case LM78_LIM_IN5:
    case LM78_LIM_IN6:
	high /= 16;		/* convert from mV */
	low  /= 16;
	if (high < 0 || high > 255 ||
	    low  < 0 || low > 255) return -EINVAL;
	if (high < low) return -EINVAL;
	LM78_set(reg,   high);
	LM78_set(reg+1, low);
	break;
#undef high
#undef low
/*--- STATUS items ---*/
    case LM78_DATA_CHASSIS:
	LM78_clear_chassis();
	break;
    case LM78_DATA_INTERRUPTS:
	LM78_flush_interrupts();
	break;

/*--- CONTROL items---*/
    case LM78_CONTROL_RESET:
	LM78_hardware_reset();
	break;
    case WB_BEEP_CTRL:
	printk("LM78: Unimplemented: Winbond beep control.\n");
	break;
    default:
	printk("Huh? Unknown item in LM78_set_item: 0x%x\n", reg);
	return -EINVAL;
    }
    LM78_flush_interrupts();
    return 0;
}

static void
LM78_get_item(int reg, void *data) {
    /*--- Read the limit from the chip -------------------------*/
    switch(reg) {
/*--- LIMIT items ---*/
#define low_limit (DATA_int(0))
#define divisor   (DATA_int(1))
    case LM78_LIM_FAN1:
    case LM78_LIM_FAN2:
    case LM78_LIM_FAN3:		/* low = divisor in this case */
	divisor   = LM78_get_fan_divisor(reg-LM78_LIM_FAN1+1);
	low_limit = LM78_count_to_speed(LM78_get(reg), divisor);
	break;
#undef low_limit
#undef divisor
#define high (DATA_int(0))
#define low  (DATA_int(1))
    case LM78_LIM_OVERTEMP:
	high = (signed char) LM78_get(reg);
        low  = (signed char) LM78_get(reg+1);
	break;
    case WB_LIM_HYST_2:
    case WB_LIM_HYST_3:
	high = (signed char) WB_get(reg+2);
	low  = (signed char) WB_get(reg);
	break;
#undef low
#undef high
#define high (DATA_int(0))
#define low  (DATA_int(1))
    case LM78_LIM_IN0:
    case LM78_LIM_IN1:
    case LM78_LIM_IN2:
    case LM78_LIM_IN3:
    case LM78_LIM_IN4:
    case LM78_LIM_IN5:
    case LM78_LIM_IN6:
	high = 16 * (int) LM78_get(reg);
	low  = 16 * (int) LM78_get(reg+1);
	break;
	
/*--- STATUS items ---*/
#undef low
#undef high
#define value (DATA_int(0))
    case LM78_DATA_FAN1:
    case LM78_DATA_FAN2:
    case LM78_DATA_FAN3:         /* low = divisor in this case */
	value = LM78_get_fan_speed(reg - LM78_DATA_FAN1+1);
	break;
    case LM78_DATA_TEMP:
        value = (signed char) LM78_get(reg);
	break;
    case WB_DATA_TEMP_2:
    case WB_DATA_TEMP_3:
	value = (signed char) WB_get(reg);
	break;
    case LM78_DATA_CHASSIS:
	value = (LM78_get_interrupts() & LM78_INT_CHASSIS) ? 1 : 0;
	break;
    case LM78_DATA_IN0:
    case LM78_DATA_IN1:
    case LM78_DATA_IN2:
    case LM78_DATA_IN3:
    case LM78_DATA_IN4:
    case LM78_DATA_IN5:
    case LM78_DATA_IN6:
        value = LM78_get(reg) * 16; /* reading are in mV */
	break;
#undef value
#define value DATA_string
    case LM78_DATA_INTERRUPTS:
	lm78_print_ints(value);
	break;
#undef value
#define value (DATA_int(0))

/*--- CONTROL items---*/
    case LM78_CONTROL_RESET:
	value = 0;
	break;
    case WB_BEEP_CTRL:
	printk("LM78: Unimplemented: Winbond beep control.\n");
	break;
    default:
	printk("Random reg get: %d\n", reg);
        value = LM78_get(reg);
	break;
    }
}

/*------------------------------------------------------------------*/
/*------------------------------------------------------------------*/
/*------------------------------------------------------------------*/
/*------------------------------------------------------------------*/
#define MAXCTL 30
static struct ctl_table lim_table[MAXCTL];
static struct ctl_table stat_table[MAXCTL];
static struct ctl_table control_table[MAXCTL];

static struct ctl_table lm78_dir[] = {
    {HM_LIMITS,  "limits", NULL, 0, 0555, lim_table},
    {HM_STATUS,  "status", NULL, 0, 0555, stat_table},
    {HM_CONTROL, "control",NULL, 0, 0555, control_table},
    {0,0,}
};

static struct ctl_table dev_dir[] = {
    {HM_LM78, "lm78", NULL, 0, 0555, lm78_dir},
    {0,0,}
};

static struct ctl_table dev_root[] = {
    {CTL_DEV, "dev", NULL, 0, 0555, dev_dir},
    {0,0,}
};

/*--------------------------------------------------------------------
 *  handlers for sysctl and /proc
 */
static int
lm78_sysctl_handler(ctl_table *table, int *name, int nlen,
		   void *oldval, size_t *oldlenp,
		   void *newval, size_t newlen, void **context) {
    int len, ulen, err=0, reg;
    void *data;
    
    MOD_INC_USE_COUNT;

    len  = *((int *)table->extra1) & SC_SIZE_MASK;
    data = kmalloc(len,GFP_KERNEL);
    if (!data) {
	MOD_DEC_USE_COUNT;
	return -ENOMEM;
    }
    reg = (int) table->extra2;
    
    /*--- read -----------------------------------------------------*/
    if (oldval && oldlenp) {
#ifdef VERSION_2_0
	ulen = get_user(oldlenp);
#else
	err = get_user(ulen, oldlenp);
	if (err) goto bail;
#endif
	if (len) {
	    LM78_get_item(reg, data);

   	    /* Shove the data back into user space */
	    if (len > ulen) len = ulen;
#ifdef VERSION_2_0
	    memcpy_tofs(oldval, data, len);
	    put_user(len, oldlenp);
#else
	    err = copy_to_user(oldval, data, len);
	    if (err) goto bail;
	    err = put_user(len,oldlenp);
	    if (err) goto bail;
#endif
	}
    }
    
    /*--- write ----------------------------------------------------*/
    if (newval && newlen > sizeof(int)) {
	memset(data, 0, len);
	if (newlen > len) newlen = len;
#if VERSION_2_0
	memcpy_fromfs(data, newval, newlen);
#else
	err = copy_from_user(data, newval, len);
	if (err) goto bail;
#endif
	err = LM78_set_item(reg, data);
    }
#ifndef VERSION_2_0
bail:
#endif
    kfree(data);
    MOD_DEC_USE_COUNT;
    if (err) return err;
    return 1;			/* Signal that we've taken care of it */
}

static int
lm78_proc_handler(ctl_table *table, int write, struct file *filp,
                  void *buffer, size_t *lenp) {
    int err, reg;
    /* Make a copy of the sysctl dummy to avoid problems with multiple
     * readers/writers since the provided functions insist on using the
     * table->data fields. */
    ctl_table dummy;

    MOD_INC_USE_COUNT;
    memcpy(&dummy, table, sizeof(ctl_table));
    dummy.maxlen = ((int)table->extra1) & SC_SIZE_MASK;
    dummy.data = kmalloc(dummy.maxlen, GFP_KERNEL);
    if (!dummy.data) {
	MOD_DEC_USE_COUNT;
	return -ENOMEM;
    }
    reg = (int) table->extra2;
    
    if (write) { /* write */
	switch(((int)table->extra1) & SC_TYPE_MASK) {
	case SC_TYPE_INT: err = proc_dointvec(&dummy, write, filp, buffer, lenp); break;
	case SC_TYPE_STR: err = proc_dostring(&dummy, write, filp, buffer, lenp); break;
	default: printk("Bad type = 0x%x\n",((int)table->extra1)); err = -EINVAL;
	}
	if (err < 0) goto bail;
	LM78_set_item(reg, dummy.data);
    } else { /* reading */
	LM78_get_item(reg, dummy.data);
	switch(((int)table->extra1) & SC_TYPE_MASK) {
	case SC_TYPE_INT: err = proc_dointvec(&dummy,write,filp,buffer,lenp); break;
	case SC_TYPE_STR: err = proc_dostring(&dummy,write,filp,buffer,lenp); break;
	default: printk("Bad type = 0x%x\n",((int)table->extra1)); err = -EINVAL;
	}
    }
bail:
    kfree(dummy.data);
    MOD_DEC_USE_COUNT;
    return err;
}

/*--------------------------------------------------------------------
 *  Funcs to help in setting up the sysctl tables
 */
static void
sysctl_add_table_item(struct ctl_table *table, int id, int reg, char *name, int mode, int flags) {
    table->ctl_name  = id;
    table->procname = name;
    table->data = 0;
    table->maxlen = 0;
    table->mode = mode;
    table->child = 0;
    table->proc_handler = lm78_proc_handler;
    table->strategy     = lm78_sysctl_handler;
    table->extra1 = (void *) flags;
    table->extra2 = (void *) reg;

    /* Always keep that list terminated. */
    table++;
    table->ctl_name = 0;
    table->procname = 0;
}

static void
sysctl_add_item(int id, int reg, char *name, int mode, int flags) {
    static struct ctl_table *ctl[4] = {lim_table, stat_table, control_table, 0};
    int c_id = (flags & SC_OP_MASK) / (((~SC_OP_MASK)+1)&SC_OP_MASK);
    sysctl_add_table_item(ctl[c_id], id, reg, name, mode, flags);
    ctl[c_id]++;
}

/*--------------------------------------------------------------------
 *  MODUDULE Stuff...
 */

#ifdef MODULE
static struct ctl_table_header *ctl_header;

int init_module(void) {
    printk("LM78: Hardware monitor driver Version %s\n"
	   "LM78: Erik Hendriks <hendriks@cesdis.gsfc.nasa.gov>\n"
	   "LM78: %s\n", LM78_VERSION, idstring);
    if (LM78_initialize()) return -EINVAL;

    setup_sysctl();

    ctl_header = register_sysctl_table(dev_root, 0);
    return 0;
}

void cleanup_module(void) {
    unregister_sysctl_table(ctl_header);
    LM78_stop();
}
#endif

#if VERSION_2_0
/* This procedure stolen from kernel/sysctl.c */
int proc_dointvec(ctl_table *table, int write, struct file *filp,
                  void *buffer, size_t *lenp) {
    int *i, vleft, first=1, len, left, neg, val;
#define TMPBUFLEN 20
    char buf[TMPBUFLEN], *p;
        
    if (!table->data || !table->maxlen || !*lenp ||
	(filp->f_pos && !write)) {
	*lenp = 0;
	return 0;
    }
    
    i = (int *) table->data;
    vleft = table->maxlen / sizeof(int);
    left = *lenp;
        
    for (; left && vleft--; i++, first=0) {
	if (write) {
	    while (left && isspace(get_user((char *) buffer)))
		left--, ((char *) buffer)++;
	    if (!left)
		break;
	    neg = 0;
	    len = left;
	    if (len > TMPBUFLEN-1)
		len = TMPBUFLEN-1;
	    memcpy_fromfs(buf, buffer, len);
	    buf[len] = 0;
	    p = buf;
	    if (*p == '-' && left > 1) {
		neg = 1;
		left--, p++;
	    }
	    if (*p < '0' || *p > '9')
		break;
	    val = simple_strtoul(p, &p, 0);
	    len = p-buf;
	    if ((len < left) && *p && !isspace(*p))
		break;
	    if (neg)
		val = -val;
	    buffer += len;
	    left -= len;
	    *i = val;
	} else {
	    p = buf;
	    if (!first)
		*p++ = '\t';
	    sprintf(p, "%d", *i);
	    len = strlen(buf);
	    if (len > left)
		len = left;
	    memcpy_tofs(buffer, buf, len);
	    left -= len;
	    buffer += len;
	}
    }
    
    if (!write && !first && left) {
	put_user('\n', (char *) buffer);
	left--, buffer++;
    }
    if (write) {
	p = (char *) buffer;
	while (left && isspace(get_user(p++)))
	    left--;
    }
    if (write && first)
	return -EINVAL;
    *lenp -= left;
    filp->f_pos += *lenp;
    return 0;
}

int proc_dostring(ctl_table *table, int write, struct file *filp,
                  void *buffer, size_t *lenp) {
        int len;
        char *p, c;
        
        if (!table->data || !table->maxlen || !*lenp ||
            (filp->f_pos && !write)) {
                *lenp = 0;
                return 0;
        }
        
        if (write) {
                len = 0;
                p = buffer;
                while (len < *lenp && 
                       (c = get_user(p++)) != 0 && c != '\n')
                        len++;
                if (len >= table->maxlen)
                        len = table->maxlen-1;
                memcpy_fromfs(table->data, buffer, len);
                ((char *) table->data)[len] = 0;
                filp->f_pos += *lenp;
        } else {
                len = strlen(table->data);
                if (len > table->maxlen)
                        len = table->maxlen;
                if (len > *lenp)
                        len = *lenp;
                if (len)
                        memcpy_tofs(buffer, table->data, len);
                if (len < *lenp) {
                        put_user('\n', ((char *) buffer) + len);
                        len++;
                }
                *lenp = len;
                filp->f_pos += len;
        }
        return 0;
}
#endif


/* JUNK YARD */
#if 0
static int tmps[50][2];
static struct ctl_table lm78_lim_table[] = {
    {LM78_LIM_IN0,      "in0", tmps[0], sizeof(int)*2, 0644, NULL, lm78_proc_lim, sysctl_setlim},
    {LM78_LIM_IN1,      "in1", tmps[1], sizeof(int)*2, 0644, NULL, lm78_proc_lim, sysctl_setlim},
    {LM78_LIM_IN2,      "in2", tmps[2], sizeof(int)*2, 0644, NULL, lm78_proc_lim, sysctl_setlim},
    {LM78_LIM_IN3,      "in3", tmps[3], sizeof(int)*2, 0644, NULL, lm78_proc_lim, sysctl_setlim},
    {LM78_LIM_IN4,      "in4", tmps[4], sizeof(int)*2, 0644, NULL, lm78_proc_lim, sysctl_setlim},
    {LM78_LIM_IN5,      "in5", tmps[5], sizeof(int)*2, 0644, NULL, lm78_proc_lim, sysctl_setlim},
    {LM78_LIM_IN6,      "in6", tmps[6], sizeof(int)*2, 0644, NULL, lm78_proc_lim, sysctl_setlim},
    {LM78_LIM_OVERTEMP, "temperature",tmps[7],sizeof(int)*2,0644,NULL,lm78_proc_lim, sysctl_setlim},
    {LM78_LIM_FAN1,     "fan1", tmps[8], sizeof(int)*2, 0644, NULL, lm78_proc_lim, sysctl_setlim},
    {LM78_LIM_FAN2,     "fan2", tmps[9], sizeof(int)*2, 0644, NULL, lm78_proc_lim, sysctl_setlim},
    {LM78_LIM_FAN3,     "fan3", tmps[10],sizeof(int)*2, 0644, NULL, lm78_proc_lim, sysctl_setlim},
    {0,0,}
};

static struct ctl_table lm78_stat_table[] = {
    {LM78_DATA_IN0,  "in0", tmps[11], sizeof(int), 0444, NULL, lm78_proc_stat, sysctl_stat},
    {LM78_DATA_IN1,  "in1", tmps[12], sizeof(int), 0444, NULL, lm78_proc_stat, sysctl_stat},
    {LM78_DATA_IN2,  "in2", tmps[13], sizeof(int), 0444, NULL, lm78_proc_stat, sysctl_stat},
    {LM78_DATA_IN3,  "in3", tmps[14], sizeof(int), 0444, NULL, lm78_proc_stat, sysctl_stat},
    {LM78_DATA_IN4,  "in4", tmps[15], sizeof(int), 0444, NULL, lm78_proc_stat, sysctl_stat},
    {LM78_DATA_IN5,  "in5", tmps[16], sizeof(int), 0444, NULL, lm78_proc_stat, sysctl_stat},
    {LM78_DATA_IN6,  "in6", tmps[17], sizeof(int), 0444, NULL, lm78_proc_stat, sysctl_stat},
    {LM78_DATA_TEMP, "temperature", tmps[18], sizeof(int), 0444, NULL, lm78_proc_stat, sysctl_stat},
    {LM78_DATA_FAN1, "fan1", tmps[19], sizeof(int), 0444, NULL, lm78_proc_stat, sysctl_stat},
    {LM78_DATA_FAN2, "fan2", tmps[20], sizeof(int), 0444, NULL, lm78_proc_stat, sysctl_stat},
    {LM78_DATA_FAN3, "fan3", tmps[21], sizeof(int), 0444, NULL, lm78_proc_stat, sysctl_stat},

    {LM78_DATA_RESET, "reset", tmps[23], sizeof(int), 0644, NULL, lm78_proc_reset, sysctl_reset},
    {0,0}
};
#endif


/*
 * Local variables:
 * c-basic-offset: 4
 * compile-command: "gcc -O -Wall -D__KERNEL__ -DMODULE -I. -c lm78.c"
 * End:
 */
